from sqlite3 import connect
import logging

logger = logging.getLogger(__name__)

class Sqlite:
    """
    This class handles everything related to the database of the app.
    """

    def __init__(self, db_file):
        # If the file doesn't exist, the connection will create it.
        self.connection = connect(db_file)

    def close_connection(self):
        """
        Close the Sqlite connection.
        """
        self.connection.close()

    def create_data_base(self):
        """
        Creates the database if it doesn't exist. Updates scheme if needed.
        """
        logger.debug("Creating database file or/and updating it's scheme")
        # file always exists since the connection creates it, so no need to check if the file exists
        cur = self.connection.cursor()

        cur.executescript("""
            create table if not exists alerts (
                "Numéro" TEXT,
                "Date" TEXT,
                "SystèmesAffectés" TEXT,
                "Résumé" TEXT,
                "ScoreNVD" TEXT,
                "Source" TEXT,
                "Details" TEXT,
                UNIQUE(Numéro)
            );
        """)

        cur.executescript("""
            create table if not exists NextAlert (
                "NextScrap" TEXT
            );
            create table if not exists NextAvis (
                "NextScrap" TEXT
            );
        """)

        if not [row[0] for row in cur.execute("SELECT * FROM NextAlert")]:
            cur.execute("INSERT INTO NextAlert VALUES (?)", ["Null"])
        if not [row[0] for row in cur.execute("SELECT * FROM NextAvis")]:
            cur.execute("INSERT INTO NextAvis VALUES (?)", ["Null"])

        self.connection.commit()

    def add_new_alert(self, number, _date, affected_systems, summary, score_nvd, source, details):
        """
        Creates or updates a line in the table "alerts".
        :param number: str
        :param _date: str
        :param affected_systems: str
        :param summary: str
        :param score_nvd: str
        :param source: str
        :param details: str
        """
        cur = self.connection.cursor()

        if number not in [row[0] for row in cur.execute("select Numéro from alerts where Numéro = ?", [number])]:
            # if the alerts isn't in tha database yet.
            cur.execute("INSERT INTO alerts VALUES (?, ?, ?, ?, ?, ?, ?)",
                        (number, _date, affected_systems, summary, score_nvd, source, details))
        else:
            # if the alert already exists in the database
            cur.execute("""UPDATE alerts SET
            Numéro = ?,
            Date = ?,
            SystèmesAffectés = ?,
            Résumé = ?,
            ScoreNVD = ?,
            Source = ?,
            Details = ?
            WHERE Numéro = ?
            """, (number, _date, affected_systems, summary, score_nvd, source, details, number))

        self.connection.commit()

    def get_all_alerts_names(self) -> [str]:
        """
        Returns a list of string containing all alert's names.
        :return: [str]
        """
        logger.debug(f"Getting every alert's names from database")
        cur = self.connection.cursor()

        res = []
        for row in cur.execute("select Numéro from alerts"):
            res.append(row[0])

        return res

    def get_alerts_newer_than(self, _date) -> [str]:
        """
        Returns a list of string containing all alert's names that are newer than the date given as argument.
        :return: [str]
        :param _date: "DD-MM-YYYY"
        """
        logger.debug(f"Getting all alert's names newer than {_date}")
        cur = self.connection.cursor()

        res = []
        for row in cur.execute("select Numéro from alerts where Date >= date(?)", ["-".join(_date.split('-')[::-1])]):
            res.append(row[0])

        return res

    def get_one_row_from_alerts(self, column, number) -> str:
        """
        Returns a string containing the content of the column related to the alert's number, both given as argument.
        :param column: str
        :param number: str
        :return: str
        """
        logger.debug(f"Getting {column} column from {number}")
        cur = self.connection.cursor()

        cur.execute("select " + column + " from alerts where Numéro = ?", [number])

        return cur.fetchall()[0][0]

    def set_next_scrap(self, table, next_scrap):
        """
        Either sets "NextAlert" or "NextAvis" token of the database for the next time.
        :param table: str - "NextAlert" or "NextAvis"
        :param next_scrap: str - CERTFR-AAAA-(ALE or AVI)-NNN
        """
        logger.debug(f"Updating Nextscrap token of {table} to {next_scrap}")
        cur = self.connection.cursor()

        cur.execute("UPDATE " + table + " SET NextScrap = ?", [next_scrap])

        self.connection.commit()

    def get_next_scrap(self, table) -> str:
        """
        Returns a string containing the value of the NextScrap token.
        :param table: str - "NextAlert" ou "NextAvis"
        :return: str
        """
        logger.debug(f"Getting Nextscrap token of {table}")

        cur = self.connection.cursor()

        cur.execute("select NextScrap from " + table)

        return cur.fetchall()[0][0]



import logging
import feedparser
import requests
import sys

import Certfrtracker.NVD as NVD
import Certfrtracker.html_parser as html_parser
from Certfrtracker.sqlite import Sqlite
from Certfrtracker.version_parser import systems_filter

logger = logging.getLogger(__name__)


def exceptions_handler(type, value, tb):
    """
    new exception handler, to pipe all unexpected errors into the logger
    """
    logger.error("Uncaught exception: {0}".format(str(value)))


# change exception handler to custom exception handler
sys.excepthook = exceptions_handler


class Report:
    """
    Plain object returned by Router class as list of Report.
    It contains all information about each alert.
    """

    def __init__(self, alert_id: str, techno: str, version: str, status: str, score: float, publish_date: str,
                 update_date: str, description: str, source: str, details: str) -> None:
        self.alert_id = alert_id  # id of the Alert           | String | CVE-2022-1234,
        # CERTFR-2022-ALE-004, CERTFR-2022-AVI-004
        self.techno = techno  # Name of the Techno        | String | ex: Apache, Postgresql, Openjdk
        self.version = version  # Version of the Techno     | String | ex: 1.2.3, 1.2, 1.2.3-beta
        self.status = status  # Nature of Alert           | String | Open or Applicable
        self.score = score  # CVSS or NVD Score         | Float  | between 1 and 10
        self.publish_date = publish_date  # Publish Date              | String | ex: "2020-06-21
        self.update_date = update_date  # Update Date               | String | ex: "2020-06-21
        self.description = description  # Alert Description         | String
        # self.source = | String | ex: url
        self.source = source  # url of the alert | string
        self.details = details


class Router:
    """
    This class creates and updates the database to the certfr.
    It also compares the technos and versions you give as argument to the database and returns it as a list of Report
    """

    def __init__(self, db_file="CertfrTracker.db"):
        self.sqliteCon = Sqlite(db_file)
        self.sqliteCon.create_data_base()

    def __del__(self):
        self.sqliteCon.close_connection()

    def update_database(self, feed_type=""):
        """
        Updates the database by scrapping the certfr, and it's RSS stream.
        :param feed_type: str="" - "NextAlert" or "NextAvis", if null exec both.
        """
        if feed_type == "NextAvis":
            rss_link = 'https://www.cert.ssi.gouv.fr/avis/feed/'
            last_scrap = self.sqliteCon.get_next_scrap(feed_type)  # get lastScrap from database
            if last_scrap == "Null":
                last_scrap = "CERTFR-2015-AVI-001"
        elif feed_type == "NextAlert":
            rss_link = 'https://www.cert.ssi.gouv.fr/alerte/feed/'
            last_scrap = self.sqliteCon.get_next_scrap(feed_type)  # get lastScrap from database
            if last_scrap == "Null":
                last_scrap = "CERTFR-2014-ALE-001"
        else:
            self.update_database("NextAvis")
            self.update_database("NextAlert")
            return

        # TODO: secure errors
        logger.info(f"Starting new HTTPS connection: {rss_link}")
        news_feed = feedparser.parse(rss_link)

        # since the order of the feed is not sorted, we have to manually detect which of theses are the latest and
        # newest alerts
        data_set = {"date": [], "link": []}
        for entry in news_feed.entries:
            data_set["date"].append(entry.published_parsed)
            data_set["link"].append(entry.link)
        latest_feed = data_set["link"][data_set["date"].index(max(data_set["date"]))].split("/")[-2]
        oldest_feed = data_set["link"][data_set["date"].index(min(data_set["date"]))].split("/")[-2]

        # if the database is not up-to-date
        if last_scrap != latest_feed:
            self.update_database_using_rss(data_set["link"])
            self.sqliteCon.set_next_scrap(feed_type, latest_feed)  # set next_scrap to database

        # if there's a hole between the oldest rss and the latest from database :
        if int(last_scrap.split('-')[1]) <= int(oldest_feed.split('-')[1]) and int(last_scrap.split('-')[-1]) < int(
                oldest_feed.split('-')[-1]):
            self.update_database_using_crawler(feed_type=feed_type, last_scrap=last_scrap, max_crawl=oldest_feed)

        logger.info(f"{feed_type} Database is up-to-date")

    def update_database_using_rss(self, urls: [str]):
        """
        Updates the database by scrapping the certfr
        :param urls: [str] - a list containing all the alerts that are going to be saved.
        """
        logger.info(f"Updating via RSS")

        for url in urls:
            # Download from URL.
            r = requests.get(url)
            logger.info(f"Starting new HTTPS connection: {url}")
            logger.info(f"received status {r.status_code} from {url}")

            # Save to database.
            logger.info(f"Saving {url} alert into database...")
            self.alerts_to_database(url.split("/")[-2], url, r.content)

        logger.info(f"RSS Update finished")

    def update_database_using_crawler(self, feed_type, last_scrap, max_crawl):
        """
        Updates the database by crawling the certfr.
        :param feed_type: str - "NextAlert" or "NextAvis"
        :param last_scrap : str - the alert to begin the crawl with
        :param max_crawl : str - the alert to stop the crawl with
        """
        logger.info(f"Updating via Crawler")

        if feed_type == "NextAvis":
            url = 'https://www.cert.ssi.gouv.fr/avis/'
        else:
            url = 'https://www.cert.ssi.gouv.fr/alerte/'

        begin_year = int(last_scrap[7:11])
        begin_alert = int(last_scrap.split('-')[-1])

        max_number = 10000
        max_year = int(max_crawl[7:11])

        for year in range(begin_year, max_year + 1):
            if year != begin_year:
                begin_alert = 1
            # if this is the last year of crawling, it sets the max value to the value of crawl
            if year == max_year:
                max_number = int(max_crawl.split('-')[-1])

            for count in range(begin_alert, max_number):
                # Download from URL.
                alert_id = 'CERTFR-' + str(year) + '-' + feed_type.upper()[4:7] + '-' + str(count).zfill(3)
                r = requests.get(url + alert_id)
                logger.info(f"Starting new HTTPS connection: {url}{alert_id}")
                logger.info(f"received status {r.status_code} from {alert_id}")

                # return the next alert to begin with for next run
                if r.status_code == 404 and year == max_year:
                    logger.info(f"Crawler update finished")
                    return
                # go to next year once the actual is done
                elif r.status_code == 404:
                    break

                # Save to database.
                logger.info(f"Saving {alert_id} alert into database...")
                self.alerts_to_database(alert_id, url + alert_id, r.content)

    def alerts_to_database(self, alert_id, source, text):
        """
        Scraps the html file in argument and insert it into the database.
        :param alert_id: str - CERTFR-AAAA-(ALE or AVI)-NNN
        :param source: str - URL of the alert
        :param text: str - html content returned by the certfr
        """
        systems_affectes = html_parser.systems_parser(text)
        _date = html_parser.date_parser(text)
        summary = html_parser.header_summary_parser(text)
        documentation_texte, documentation_liens = html_parser.documentation_parser(text)

        # NVD score
        CVE = NVD.checkForCVE(documentation_liens, documentation_texte)

        score = 0.0
        if CVE != "":  # if a CVE has been detected it will replace 0.0 by the actual severity
            logger.debug(f"CVE \"{CVE}\" found for {alert_id} alert")
            score = NVD.getNVDScore(CVE)

        # get details link
        details = html_parser.define_details(score, CVE, documentation_liens)

        self.sqliteCon.add_new_alert(alert_id, _date, systems_affectes, summary, score, source, details)

    def compare_inventory_with_alerts(self, technos: [str], versions: [str], dates: [str]) -> [Report]:
        """
        returns a list of Report by comparing the inventory in entry with the database.
        :param technos: [str]
        :param versions: [str]
        :param dates: [str]
        :return: [Report]
        """
        reports = []

        for techno, version, _date, in zip(technos, versions, dates):
            reports += self.compare_one_techno_with_alerts(techno, version, _date)

        return reports

    def compare_one_techno_with_alerts(self, techno: str, version: str, _date: str) -> [Report]:
        """
        returns a list of Report by comparing the single techno in entry with the database.
        :param techno: str
        :param version: str
        :param _date: str
        :return: [Report]
        """
        logger.info(f"Comparing {techno} {version} with the database...")
        reports = []
        techno = techno.lower()
        if _date == "":
            _date = "01-01-2014"

        for alert in self.sqliteCon.get_alerts_newer_than(_date):
            for line in self.sqliteCon.get_one_row_from_alerts('SystèmesAffectés', alert).split("|"):
                result = systems_filter(techno, version, line)
                if result is not None:
                    score_nvd = self.sqliteCon.get_one_row_from_alerts('ScoreNVD', alert)
                    summary = self.sqliteCon.get_one_row_from_alerts('Résumé', alert)
                    release_date = self.sqliteCon.get_one_row_from_alerts('Date', alert)
                    source = self.sqliteCon.get_one_row_from_alerts('Source', alert)
                    details = self.sqliteCon.get_one_row_from_alerts('Details', alert)

                    report = Report(alert_id=alert, techno=techno, version=version, status=result,
                                    score=float(score_nvd), update_date=release_date, publish_date=release_date,
                                    description=summary, source=source, details=details)

                    reports.append(report)

        # to remove duplicated from list
        no_duplicates = []

        logger.debug(f"Removing duplicates in reports")
        for report in reports:
            token = True
            for index, no_duplicate in enumerate(no_duplicates):
                # check if alert_id already exists in "no_duplicate"
                if report.alert_id == no_duplicate.alert_id:
                    token = False
                    # check if the existing alert is "Applicable" or "Open"
                    # if the status of the existing alert is lower, it will replace it by the higher report
                    if report.status[0] == "A" and no_duplicate.status[0] == "O":
                        no_duplicates[index] = report

            if token:
                no_duplicates.append(report)

        return no_duplicates



from re import search
from requests import get
from json import loads, JSONDecodeError
import logging

logger = logging.getLogger(__name__)

def checkForCVE(documentation_links, documentation_text) -> str:
    """
    Returns a string containing either a str or None.
    Searches in the documentation given as argument an occurrence of a CVE.
    :param documentation_links: [str]
    :param documentation_text: [str]
    :return: str
    """
    logger.debug(f"Detecting CVE")

    for link in documentation_links:
        res = search(r'CVE-\d{4}-\d{4}', link)
        if res is not None:
            return res.group(0)

    for line in documentation_text:
        res = search(r'CVE-\d{4}-\d{4}', line)
        if res is not None:
            return res.group(0)

    return ""


def getNVDScore(CVE) -> float:
    """
    Returns a float containing the severity score of the CVE, established by NVD.
    :param CVE: str
    :return: float
    """
    logger.debug(f"Getting NVD score for {CVE}")

    url = "https://services.nvd.nist.gov/rest/json/cve/1.0/" + CVE
    r = get(url)

    logger.info(f"Starting new HTTPS connection: {url}")
    logger.info(f"received status {r.status_code} from {CVE}")

    try:
        json_data = loads(r.text)
    except JSONDecodeError:
        logger.warning(f"NVD API is not responding for {url}")
        return 0.0

    if not PageExists(json_data):
        logger.debug(f"NVD page for {CVE} does not exist")
        return 0.0

    parsed_data = json_data["result"]['CVE_Items'][0]['impact']
    try:
        if parsed_data.get('baseMetricV3') is None:
            score = str(parsed_data['baseMetricV2']["cvssV2"]['baseScore'])

            return float(score)
        else:
            parsed_data = parsed_data['baseMetricV3']["cvssV3"]
            BaseScore = str(parsed_data['baseScore']) + " " + parsed_data['baseSeverity']

            return float(BaseScore.split()[0])
    except KeyError:
        return 0.0


def PageExists(json_data) -> bool:
    """
    Returns a boolean depending on the json_data. Since NVD's API doesn't respond with 404 status code when a CVE
    page doesn't exist. We need to inspect the content of the json response.
    :param json_data: str
    :return: int: 1 or 0
    """
    if not json_data.get("message") is None:
        if json_data['message'].startswith("Unable to find vuln"):
            return False

    return True



from bs4 import BeautifulSoup
from re import sub
import logging

logger = logging.getLogger(__name__)

def systems_parser(fileContent) -> [str]:
    """
    returns a string containing the content under the headers. separated by "|" for each line.
    :param fileContent: str - html content
    :return: [str]
    """
    logging.debug(f"parsing \"Systèmes affectés\"")
    content = ""

    soup = BeautifulSoup(fileContent, 'html.parser')
    for header_title in soup.find_all('h2'):
        if str(header_title) in ('<h2>Systèmes affectés</h2>', '<h2>Systèmes affecté(s)</h2>'):
            # this "while" statement is the best way to get safely and only the content we need.
            cache_header = header_title
            while True:
                try:
                    cache_header = cache_header.next_sibling
                # if there isn't any "<ul>" tag in header_title, it will get the closest tag from header_title and
                # insert it in database
                except AttributeError:
                    content = sub('<.*?>', '', str(header_title.next_sibling))
                    break
                if str(cache_header).startswith("<ul>"):
                    content = sub('</li> <li>', '|', str(cache_header)[9:-11])
                    break

    return content.replace(u'\xa0', u' ')


def documentation_parser(fileContent) -> ([str], [str]):
    """
    returns a tuple containing 2 array of string containing the content under the header "documentation". each index
    containing a line.
    :param fileContent: str - html content
    :return: ([str], [str])
    """
    logging.debug(f"parsing \"Documentation\"")
    titles = []
    links = []

    soup = BeautifulSoup(fileContent, 'html.parser')
    for header_title in soup.find_all('h2'):
        if str(header_title) == '<h2>Documentation</h2>':
            # the triple "next_element" might looks ugly,
            # but it always get the content of the headers we need to filter
            temp_header = header_title.next_element.next_element.next_element
            for title in temp_header.select('ul > li'):
                titles.append(str(title.next_element).replace(u'\xa0', u' '))
            for link in temp_header.select('a'):
                links.append(str(link.next_element).replace(u'\xa0', u' '))

    return titles, links


def header_summary_parser(fileContent) -> [str]:
    """
    returns an array of string containing the content under the headers. each index containing a line.
    :param fileContent: str - html content
    :return: [str]
    """
    logging.debug(f"parsing \"Résumé\"")
    content = ""

    soup = BeautifulSoup(fileContent, 'html.parser')
    for header_title in soup.find_all('h2'):
        if str(header_title) == '<h2>Résumé</h2>':
            # the triple "next_element" might look ugly,
            # but it always gets the content of the headers we need to filter
            content = sub('<.*?>', '', str(header_title.next_element.next_element.next_element))

    return content.replace(u'\xa0', u' ')


def date_parser(file_content) -> str:
    """
    Returns a string containing a date (YYYY-MM-DD).
    :param file_content: str - html content
    :return: str
    """
    logging.debug(f"parsing \"Date de la dernière version\"")

    soup = BeautifulSoup(file_content, 'html.parser')
    table = soup.find("table", class_="table table-condensed")
    allValues = table.find_all('td')
    for index, text in enumerate(allValues):
        if "Date de la dernière version" in text:
            months = ['janvier', 'février', 'mars', 'avril', 'mai', 'juin', 'juillet', 'août', 'septembre', 'octobre',
                      'novembre', 'décembre']
            date = sub('<.*?>', '', str(allValues[index + 1])).split()[:3]
            date[0] = date[0].zfill(2)
            for i, month in enumerate(months):
                if month == date[1]:
                    date[1] = str(i + 1).zfill(2)
                    return '-'.join(date[::-1])


def define_details(score, CVE, documentation_liens):
    """
    Returns a string containing a URL to an external resource about the CVE.
    If there isn't any CVE it will return the first link of "documentation_liens" as it should the most useful.
    :param score: float
    :param CVE: str - CVE-YYYY-NNNN
    :param documentation_liens: [str]
    """
    logging.debug(f"parsing \"all links to get the most interesting\"")

    if score > 0.0:
        return ("https://nvd.nist.gov/vuln/detail/" + CVE).replace(u'\xa0', u' ')
    if CVE != "":
        return ("http://cve.mitre.org/cgi-bin/cvename.cgi?name=" + CVE).replace(u'\xa0', u' ')
    if len(documentation_liens) == 0:
        return "No further details".replace(u'\xa0', u' ')
    else:
        return documentation_liens[0].replace(u'\xa0', u' ')



from re import sub
import logging

logger = logging.getLogger(__name__)

def systems_filter(techno, json_version, line):
    """
    Returns a string containing either "Open" or "Applicable" depending on the level of confidence.
    checks if techno and version given in argument matches with the line argument.
    :param techno: str
    :param json_version: str
    :param line: str
    :return: str or None
    """
    logger.debug(f"Comparing \"{techno}\" \"{json_version}\" with \"{line}\"")

    cleared_line = sub(r"[^a-zA-Z0-9-éàèêûïöë.]", " ", line.lower())
    if cleared_line[-1:] == ".":  # sometimes the line could finish by a ".", so it can confuse the parsers
        cleared_line = cleared_line[:-1]

    if techno in cleared_line.split():
        all_check = [
            perfect_match(cleared_line, json_version),
            double_range_no_x(cleared_line, json_version),
            range_x(cleared_line, json_version),
            double_range_x(cleared_line, json_version),
            double_range_plus_extra(cleared_line, json_version),
        ]

        for single_check in all_check:
            if single_check:
                return "Applicable"

        return "Open"

    logger.debug(f"\"{techno}\" has not been found in \"{line}\"")


def range_x(cleared_line, json_version) -> bool:
    """
    Returns a Bool by comparing the line with the version.
    "True" is it match with the pattern it analyses.
    :param cleared_line: str - eg : Apache Log4j versions 2.x (versions obsolètes)
    :param json_version: str
    :return: bool
    """

    if ".x" in cleared_line and not all(x in cleared_line for x in ["antérieures", "à"]):
        logger.debug(f"Seems to match x_range filter")

        version = ""
        for word in cleared_line.split():
            if ".x" in word:
                version = word

        # quit function if nothing detected
        if "" == version:
            return False

        dot_number = len(version.split(".")) - 1
        left_of_value = '.'.join(version.split(".")[0:-1])
        json_left_of_value = '.'.join(json_version.split(".")[0:dot_number])

        if left_of_value == json_left_of_value:
            return True
        elif len(version) > len(json_version) and version[0:len(json_version)] == json_version:
            return True

    return False


def double_range_no_x(cleared_line, json_version) -> bool:
    """
    Returns a Bool by comparing the line with the version.
    "True" is it match with the pattern it analyses.
    :param cleared_line: str - eg : vCenter Server versions 7.0 antérieures à 7.0 U2b
    :param json_version: str
    :return: bool
    """
    if (("antérieures à" in cleared_line) or ("à" in cleared_line and not "antérieures à" in cleared_line)) and (
            not ".x" in cleared_line):
        logger.debug(f"Seems to match double_range_no_x filter")

        version1 = ""
        version2 = ""
        for word in cleared_line.split():
            if "." in word and version1 == "":
                version1 = word
            if "." in word and version2 == "" and word != version1 and version1 != "":
                version2 = word

        # quit function if nothing detected
        if "" in (version1, version2):
            return False

        json_version_as_array = Private.version_to_int_array(json_version)
        version1_as_array = Private.version_to_int_array(version1)
        version2_as_array = Private.version_to_int_array(version2)

        return Private.double_range_comparator(version1_as_array, version2_as_array, json_version_as_array, False, 0)

    return False


def double_range_x(cleared_line, json_version) -> bool:
    """
    Returns a Bool by comparing the line with the version.
    "True" is it match with the pattern it analyses.
    :param cleared_line: str - eg : Cloud Foundation (vCenter Server) versions 3.x antérieures à 3.10.2.1
    :param json_version: str
    :return: bool
    """
    if (("antérieures à" in cleared_line) or ("à" in cleared_line and not "antérieures à" in cleared_line)) and (
            ".x" in cleared_line):
        logger.debug(f"Seems to match double_range_x filter")

        version1 = ""
        version2 = ""
        for word in cleared_line.split():
            if ".x" in word and version1 == "":
                version1 = word
            if "." in word and version2 == "" and word != version1 and version1 != "":
                version2 = word

        # quit function if nothing detected
        if "" in (version1, version2):
            return False

        dot_number = len(version1.split(".")) - 1
        left_of_value = '.'.join(version1.split(".")[0:-1])
        v1_sanitized = left_of_value + (".0" * (len(version2.split(".")) - dot_number))

        json_version_as_array = Private.version_to_int_array(json_version)
        version1_as_array = Private.version_to_int_array(v1_sanitized)
        version2_as_array = Private.version_to_int_array(version2)

        return Private.double_range_comparator(version1_as_array, version2_as_array, json_version_as_array, True, 0)

    return False


def perfect_match(cleared_line, json_version) -> bool:
    """
    Returns a Bool by comparing the line with the version.
    "True" is it match with the pattern it analyses.
    :param cleared_line: str - eg : Apache Log4j versions 2.16.0 et 2.12.2 (java 7)
    :param json_version: str
    :return: bool
    """
    if all(x + ' ' in cleared_line for x in json_version):
        logger.debug(f"Seems to match perfect_match filter")
        return True
    return False


def double_range_plus_extra(cleared_line, json_version) -> bool:
    # apache tomcat versions 8.5.50 à 8.5.81 antérieures à 8.5.82
    # PostgreSQL JDBC versions 42.3.x et 42.4.x antérieures à 42.4.1
    """
    Returns a Bool by comparing the line with the version.
    "True" is it match with the pattern it analyses.
    :param cleared_line: str - eg : apache tomcat versions 8.5.50 à 8.5.81 antérieures à 8.5.82
    :param json_version: str
    :return: bool
    """
    if len(cleared_line.split(" et ")) == 2 and len(cleared_line.split(" antérieures à ")) == 2:
        line = cleared_line.split(" et ")
    elif len(cleared_line.split(" à ")) == 3 and len(cleared_line.split(" antérieures ")) == 2:
        line = cleared_line.split(" a ")
    else:
        return False

    logger.debug(f"Seems to match double_range_plus_extra filter. It will be redirect to another filter.")

    # If the wording matches the filter, it will reparse "cleared_line" to be tested.
    version1 = line[0].split()[-1]
    version2 = line[-1].split()[-1]
    if ".x" in cleared_line:
        double_range_x(version1 + " antérieures à " + version2, json_version)
    else:
        double_range_no_x(version1 + " antérieures à " + version2, json_version)


class Private:
    """
    Private class inside of module "version_parser". this class doesn't need to be imported.
    """

    @staticmethod
    def double_range_comparator(version1, version2, json_version, x_range, recursion_level) -> bool:
        """
        Return a bool by comparing if version1 <= json_version <= version2 via recursion.
        :param version1: [int]
        :param version2: [int]
        :param json_version: [int]
        :param x_range: bool
        :param recursion_level: int
        :return: bool
        """
        if recursion_level in (len(version1), len(version2), len(json_version)):
            return True
        if version1[recursion_level] <= json_version[recursion_level] < version2[recursion_level] and x_range:
            return True
        if version1[recursion_level] < json_version[recursion_level] < version2[recursion_level]:
            return True
        elif json_version[recursion_level] in [version1[recursion_level], version2[recursion_level]]:
            return Private.double_range_comparator(version1, version2, json_version, x_range, recursion_level + 1)
        else:
            return False

    @staticmethod
    def version_to_int_array(version) -> [int]:
        """
        Return an array of int containing the version given in argument converted.
        :param version: str - eg : "3.4.0"
        :return: [int] - eg : [3, 4, 0]
        """
        version = version.split(".")

        for index, value in enumerate(version):
            for char in value:  # if non digit character follows the version
                if not char.isdigit():
                    version[index] = value.split(char)[0]
                    break
            version[index] += "0" if version[index] == "" else ""

        map_version = map(int, version)
        return list(map_version)
